/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

"use strict";

const EXPORTED_SYMBOLS = ["BrowserWindow", // new-window.js
                         ];

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");
Components.utils.import("resource://multifox-modules/new-window.js");

console.log("===>LOADING main.js");

var m_tld = null;

// tabbrowser attributes:
// "multifox-tabbrowser-last-tab"

// tab attributes:
// "multifox-tab-login-tld"
// "multifox-login-submit-user"
// "multifox-tab-error" - sandbox error, page not supported

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var m_runner = null;

const BrowserWindow = {

  register: function(win) {
    console.log("BrowserWindow.register");

    if (m_runner === null) {
      // first multifox tab!
      m_runner = new MultifoxRunner();
    }

    win.document.documentElement.setAttribute("multifox-window-uninitialized", "true");
    win.addEventListener(m_runner.eventSentByContent, onContentEvent, false, true);

    // some MultifoxContentEvent_* listeners are not called when
    // there are "unload" listeners with useCapture=true. o_O
    // But they are called if event listener is an anonymous function.
    win.addEventListener("unload", onUnloadChromeWindow, false);
    win.addEventListener("mousedown", RedirDetector.onMouseDown, false); // command/click listeners can be called after network request
    win.addEventListener("keydown", RedirDetector.onKey, false);

    var tabbrowser = win.getBrowser();
    tabbrowser.addEventListener("pageshow", updateNonNetworkDocuments, false);

    var container = tabbrowser.tabContainer;
    container.addEventListener("TabSelect", TabContainerEvents, false);
    container.addEventListener("TabClose", TabContainerEvents, false);
    container.addEventListener("SSTabRestoring", TabContainerEvents, false);

    // restore icon after toolbar customization
    var toolbox = win.document.getElementById("navigator-toolbox");
    toolbox.addEventListener("DOMNodeInserted", customizeToolbar, false);
  },


  unregister: function(win) {
    console.log("BrowserWindow.unregister");

    win.removeEventListener(m_runner.eventSentByContent, onContentEvent, false);
    win.removeEventListener("mousedown", RedirDetector.onMouseDown, false);
    win.removeEventListener("keydown", RedirDetector.onKey, false);

    var toolbox = win.document.getElementById("navigator-toolbox");
    toolbox.removeEventListener("DOMNodeInserted", customizeToolbar, false);

    var tabbrowser = win.getBrowser();
    tabbrowser.removeEventListener("pageshow", updateNonNetworkDocuments, false);

    var container = tabbrowser.tabContainer;
    container.removeEventListener("TabSelect", TabContainerEvents, false);
    container.removeEventListener("TabClose", TabContainerEvents, false);
    container.removeEventListener("SSTabRestoring", TabContainerEvents, false);

    // TODO last window?
    /*
    if () {
      m_runner.shutdown();
      m_runner = null;
    }
    */
    win.removeEventListener("unload", onUnloadChromeWindow, false);
  }
};


function onUnloadChromeWindow(evt) {
  var win = evt.currentTarget;
  BrowserWindow.unregister(win);
}


const TabContainerEvents = {
  handleEvent: function(evt) {
    try {
      this[evt.type](evt);
    } catch (ex) {
      Components.utils.reportError(ex);
      throw new Error("TabContainerEvents exception " + ex);
    }
  },

  TabClose: function(evt) {
    MoveTabWindows.tabCloseSaveId(evt.originalTarget);
  },

  TabSelect: function(evt) {
    var tab = evt.originalTarget;
    MoveTabWindows.tabSelectDetectMove(tab); // moved tab: set id
    updateUI(tab, true);
  },

  SSTabRestoring: function(evt) {
    var tab = evt.originalTarget;
    updateUI(tab, true);
  }
};


// first run?
function migrating() {
  var prefName = "extensions.multifox@hultmann.currentVersion";
  var prefs = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefBranch2);
  var ver = prefs.prefHasUserValue(prefName) ? prefs.getCharPref(prefName) : "";
  if (ver === "2.0b1") {
    return;
  }
  prefs.setCharPref(prefName, "2.0b1");

  // remove Multifox 1.x cookies
  var profileId = 2;
  var all;
  do {
    all = removeTldData_cookies("multifox-profile-" + profileId);
    console.log("Migrating: removing cookies", profileId, ":", all.length);
    profileId++;
  } while (all.lenght > 0); // TODO ir até 10 no min
}


function customizeToolbar(evt) {
  var node = evt.target;
  if ((node.id === "urlbar-container") && (node.parentNode.tagName === "toolbar")) {
    updateUI(node.ownerDocument.defaultView.getBrowser().selectedTab, true);
  }
}


function onContentEvent(evt) {
  evt.stopPropagation();

  var obj = JSON.parse(evt.data);
  var contentDoc = evt.target;
  var rv;

  switch (obj.from) {
    case "cookie":
      rv = documentCookie(obj, contentDoc);
      break;
    case "localStorage":
      rv = windowLocalStorage(obj, contentDoc);
      break;
    case "error":
      showError(contentDoc.defaultView, obj.cmd, "-");
      break;
    default:
      throw new Error("onContentEvent: " + obj.from);
  }

  if (rv === undefined) {
    // no response
    return;
  }

  // send data to content
  var evt2 = contentDoc.createEvent("MessageEvent");
  evt2.initMessageEvent(m_runner.eventSentByChrome, false, false, rv, null, null, null);
  var success = contentDoc.dispatchEvent(evt2);
}


function MultifoxRunner() {
  console.log("MultifoxRunner");
  m_tld = Cc["@mozilla.org/network/effective-tld-service;1"].getService(Ci.nsIEffectiveTLDService);
  //m_cookieMgr = Cc["@mozilla.org/cookiemanager;1"].getService(Ci.nsICookieManager2); // TODO


  this._sentByChrome  = "multifox-chrome_event-"  + Math.random().toString(36).substr(2);
  this._sentByContent = "multifox-content_event-" + Math.random().toString(36).substr(2);
  this._inject = new DocStartScriptInjection();

  var obs = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
  obs.addObserver(SubmitObserver, "earlyformsubmit", false);
  obs.addObserver(LoginData.onCookieRejected, "cookie-rejected", false);
  obs.addObserver(LoginData.onCookieChanged, "cookie-changed", false);

  Cookies.start();
  util.networkListeners.enable(httpListeners.request, httpListeners.response);

  migrating();
}


MultifoxRunner.prototype = {
  get eventSentByChrome() {
    return this._sentByChrome;
  },

  get eventSentByContent() {
    return this._sentByContent;
  },

  shutdown: function() {
    console.log("MultifoxRunner.shutdown");
    util.networkListeners.disable();
    this._inject.stop();
    Cookies.stop();
    m_tld = null;
    LoginData.shutdown();

    var obs = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
    obs.removeObserver(SubmitObserver, "earlyformsubmit");
    obs.removeObserver(LoginData.onCookieRejected, "cookie-rejected");
    obs.removeObserver(LoginData.onCookieChanged, "cookie-changed");
  }
};


function showError(contentWin, notSupportedFeature, details) {
  var msg = [];
  msg.push("ERROR=" + notSupportedFeature);
  msg.push(details);
  if (contentWin.document) {
    msg.push("location=" + contentWin.document.location);
    if (contentWin.document.documentURIObject) {
      msg.push("uri=     " + contentWin.document.documentURIObject.spec);
    }
  }
  msg.push("title=[" + contentWin.document.title + "]");
  console.log(msg.join("\n"));

  var tab = WindowParents.getContainerElement(contentWin);
  tab.setAttribute("multifox-tab-error", notSupportedFeature);
  updateUI(tab, true);
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/*
attr used:
multifox-invalidate-icon
multifox-tab-login-tld
multifox-tab-login-user16
multifox-tab-error
multifox-logging-in
multifox-tab-current-tld

*/

//var IconBadge = {


function updateUI(tab, force) { // force = optional param
  if (tab.hasAttribute("selected") === false) {
    return;
  }

  if (tab.hasAttribute("multifox-invalidate-icon")) {
    tab.removeAttribute("multifox-invalidate-icon");
  } else {
    if (force !== true) {
      return;
    }
  }

  var doc = tab.ownerDocument;
  var iconContainer = getIconContainer(doc);

  if (tab.hasAttribute("multifox-tab-login-tld") === false) {
    // remove badge
    if (iconContainer !== null) {
      iconContainer.parentNode.removeChild(iconContainer);
    }
    return;
  }

  if (iconContainer === null) {
    var ref = doc.getElementById("urlbar-icons");
    iconContainer = ref.insertBefore(doc.createElement("hbox"), ref.firstChild);
    iconContainer.setAttribute("hidden", "true");
    iconContainer.setAttribute("id", "multifox-icon");
    var win = doc.defaultView;
    var delay = getTabs(win.getBrowser()).length > 1 ? 25 : 200;
    win.setTimeout(initIcon, delay, doc); // open window => greater delay
  } else {
    LoginData.setTabAsDefaultLogin(tab); // TODO it will call _ensureValid, make sure we don't block TabSelect
    if (iconContainer.firstChild !== null) { // waiting initIcon?
      updateIconCore(tab, iconContainer);
    }
  }
}


function invalidateUI(tab) {
  if (tab.hasAttribute("selected")) {
    tab.setAttribute("multifox-invalidate-icon", "true");
  }
}


function updateIconCore(tab, iconContainer) {
  updateIconUserName(tab);
  updateIconError(tab, iconContainer);
}


function updateIconUserName(tab) {
  var tabLogin = TabLogin.getFromTab(tab);
  var user;
  if (tabLogin.loginUser16 === TabLogin.NewAccount) {
    user = util.getText("icon.user.new.label");
  } else {
    user = hexToString(tabLogin.loginUser16); // BUG sometimes is null
  }

  console.log("updateIconUserName", user, tabLogin.loginUser16);
  getIconLabel(tab.ownerDocument).setAttribute("value", user);
}


function updateIconError(tab, iconContainer) {
  var currentError = iconContainer.getAttribute("current-error");
  var newError = tab.getAttribute("multifox-tab-error");
  if (currentError === newError) {
    return;
  }

  iconContainer.setAttribute("current-error", newError);
  var doc = tab.ownerDocument;
  var statIcon = getStatIconContainer(doc);
  while (statIcon.firstChild) {
    statIcon.removeChild(statIcon.firstChild);
  }

  if (newError.length === 0) {
    statIcon.setAttribute("hidden", "true");
    return;
  }

  var img = statIcon.appendChild(doc.createElement("image"));
  img.setAttribute("src", "chrome://global/skin/icons/warning-16.png"); // ubuntu: 22x22
  img.setAttribute("width", "16");
  img.setAttribute("height", "16");
  img.style.margin = "0 -4px 0 7px";
  statIcon.removeAttribute("hidden");
}


// <hbox align="center" id="multifox-icon">
//   <box>
//     <hbox align="center">
//       <hbox id="multifox-icon-stat-icon">
//         <image src="warning.png"/>
//       </hbox>
//       <label value="username"/>
//     </hbox>

function getIconContainer(doc) {
  return doc.getElementById("multifox-icon");
}

function getIconLabel(doc) {
  var node = getIconContainer(doc);
  console.assert(node !== null, "getIconContainer=null");
  return node.firstChild.firstChild.children[1];
}

function getStatIconContainer(doc) {
  return doc.getElementById("multifox-icon-stat-icon");
}


function initIcon(doc) {
  var tab = doc.defaultView.getBrowser().selectedTab;
  LoginData.setTabAsDefaultLogin(tab);

  var iconContainer = getIconContainer(doc);
  if (iconContainer === null) {
    // e.g. user changed tab!
    console.log("initIcon ignored! " + iconContainer);
    return;
  }

  console.assert(iconContainer.children.length === 0, "iconContainer has children");
  iconContainer.setAttribute("align", "center");

  var icon = iconContainer.appendChild(doc.createElement("box"));
  var labelContainer = icon.appendChild(doc.createElement("hbox"));
  var stat = labelContainer.appendChild(doc.createElement("hbox"));
  var labelId = labelContainer.appendChild(doc.createElement("label"));

  icon.setAttribute("tooltiptext", "Multifox (BETA)");
  labelContainer.setAttribute("align", "center");
  stat.setAttribute("hidden", "true");
  stat.setAttribute("id", "multifox-icon-stat-icon");

  var styleIcon = iconContainer.style;
  styleIcon.padding = "0";
  styleIcon.margin = "0 2px 0 1px";
  styleIcon.background = "transparent repeat-x";
  styleIcon.MozBorderRadius = "2px"; // Gecko 1.9.2
  styleIcon.borderRadius = "2px";
  styleIcon.minHeight = "16px";
  styleIcon.overflow = "hidden";

  if (tab.hasAttribute("multifox-logging-in")) {
    tab.removeAttribute("multifox-logging-in");
    styleIcon.opacity = "0";
    doc.defaultView.setTimeout(function() {
      styleIcon.opacity = "1";
    }, 50);
    styleIcon.MozTransition = "opacity .3s ease";
  } else {
    styleIcon.opacity = "1";
    styleIcon.MozTransition = "";
  }

  styleIcon.border = "1px solid #ccc";
  styleIcon.MozBorderRadius = "2px";

  var styleLabel = labelId.style;
  styleLabel.textRendering = "optimizelegibility";
  styleLabel.padding = "0px 9px";
  styleLabel.margin = "-1px 0px";
  styleLabel.color = "#777";

  initIconNormal(doc);
  updateIconCore(tab, iconContainer);
  iconContainer.removeAttribute("hidden");
}


function initIconNormal(doc) {
  var statIcon = getStatIconContainer(doc);
  var iconContainer = getIconContainer(doc);
  var lab = getIconLabel(doc).style;

  doc.defaultView.setTimeout(function() {
    // workaround
    // otherwise, mousedown on icon will close and open popup, even with stopPropagation
    statIcon.addEventListener("mousedown", showMsgPanel, false);
    iconContainer.addEventListener("mousedown", showMenuPopup, false);
  }, 0);
  iconContainer.addEventListener("mouseover", onIconHover, false);
  iconContainer.addEventListener("mouseout", onIconHover, false);

  statIcon.style.opacity = "1";
  lab.opacity = "1";
  defaultStyle(iconContainer.style, lab);
}


function initIconPressed(doc, popupId) {
  var iconContainer = getIconContainer(doc);
  iconContainer.removeEventListener("mouseover", onIconHover, false);
  iconContainer.removeEventListener("mouseout", onIconHover, false);

  var statIcon = getStatIconContainer(doc);
  var lab = getIconLabel(doc).style;
  statIcon.removeEventListener("mousedown", showMsgPanel, false);
  switch (popupId) {
    case "menu":
      statIcon.style.opacity = ".5";
      iconContainer.style.backgroundImage = "-moz-linear-gradient(#ddd, #999)";
      lab.textShadow = "0 1px 0 white, 0 1px 5px white, 3px 0 5px white, 0 -1px 5px white, -3px 0 5px white";
      break;
    case "msg":
      lab.opacity = ".5";
      break;
  }
}


function initIconHover(doc, mouseEvent) {
  var iconContainer = getIconContainer(doc);
  var lab = getIconLabel(doc).style;

  switch (mouseEvent) {
    case "mouseover":
      iconContainer.style.backgroundImage = "-moz-linear-gradient(#fff, #eee)";
      lab.textShadow = "0 1px 0 white, 0 1px 5px white, 3px 0 5px white, 0 -1px 5px white, -3px 0 5px white";
      break;
    case "mouseout":
      defaultStyle(iconContainer.style, lab);
      break;
  }
}


function defaultStyle(styleIcon, lab) {
  styleIcon.backgroundImage = "-moz-linear-gradient(#fff, #ddd)";
  lab.textShadow = "0 1px 0 white";
}


function onIconHover(evt) {
  initIconHover(evt.target.ownerDocument, evt.type);
}


function isMsgPopup(evt) {
  return evt.target.localName === "image";
}


function showMsgPanel(evt) {
  if (isMsgPopup(evt) === false) {
    return;
  }

  var doc = evt.target.ownerDocument;
  initIconPressed(doc, "msg");

  var ns = {};
  var subscript = Cc["@mozilla.org/moz/jssubscript-loader;1"].getService(Ci.mozIJSSubScriptLoader);
  subscript.loadSubScript("resource://multifox-modules/popup.js", ns); // TODO move to /subscripts
  var panel = ns.createMsgPanel(doc);

  panel.addEventListener("popuphidden", function(evt) {
    initIconNormal(doc);
    // remove error msg
    var tab = doc.defaultView.getBrowser().selectedTab;
    tab.removeAttribute("multifox-tab-error");
    updateUI(tab, true);
  }, false);

  panel.openPopup(getStatIconContainer(doc), "after_start", 0, 1);
}


function showMenuPopup(evt) {
  if (isMsgPopup(evt)) {
    return;
  }

  var doc = evt.target.ownerDocument;
  initIconPressed(doc, "menu");

  var menu = doc.getElementById("mainPopupSet")
                .appendChild(doc.createElement("menupopup"));

  var ns = {};
  var subscript = Cc["@mozilla.org/moz/jssubscript-loader;1"].getService(Ci.mozIJSSubScriptLoader);
  subscript.loadSubScript("resource://multifox-modules/popup.js", ns);
  ns.createLoginsMenu(menu, function() {initIconNormal(doc);});

  menu.openPopup(getIconContainer(doc), "after_end", 0, 1);
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2010
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


// tab moved (between windows): TabClose + TabSelect
// workaround: check call stack to detect d&d
const StackAnalyzer = {
  swapBrowsersOnTabSelect: function() {
    for (var s = Components.stack; s; s = s.caller) {
      switch (s.name) {
        case "swapBrowsersAndCloseOther":
        case "onxbldrop": //fx4
        case "_onDrop":   //fx36
          return true;
      }
    }
    return false;
  }
};


// keep multifox id when tab is moved between windows
var MoveTabWindows = {
  _lastClosedTab_domain: null,
  _lastClosedTab_user:   null,
  _remainingTabSelect:   0,

  tabCloseSaveId: function(tab) {
    if (tab.hasAttribute("multifox-tab-login-tld")) {
      var tabLogin = TabLogin.getFromTab(tab);
      this._lastClosedTab_domain = tabLogin.loginTld;
      this._lastClosedTab_user   = tabLogin.loginUser16;
      this._remainingTabSelect = 2;
    } else {
      this._remainingTabSelect = 0;
    }
  },

  tabSelectDetectMove: function(tab) {
    if (this._remainingTabSelect > 0) {
      this._remainingTabSelect--;
      if (StackAnalyzer.swapBrowsersOnTabSelect()) {
        var tld = this._lastClosedTab_domain;
        var usr = this._lastClosedTab_user;
        TabLogin.setLogin({
          tabElement:  tab,
          loginUser16: usr,
          loginTld:    tld
        });
        this._remainingTabSelect = 0;
        console.log('tabSelectDetectMove');
      }
    }
  }
};
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// Add hooks to documents (cookie, localStorage, ...)

function DocStartScriptInjection() {
  var is192 = this._is192();
  this._loader = new ScriptSourceLoader(is192);
  var obs = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
  obs.addObserver(this, this._getTopic(is192), false);
}

DocStartScriptInjection.prototype = {
  stop: function() {
    var obs = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
    obs.removeObserver(this, this._getTopic(this._is192()));
    delete this._loader;
  },

  _is192: function() {
    var info = Cc["@mozilla.org/xre/app-info;1"].getService(Ci.nsIXULAppInfo);
    return info.platformVersion.indexOf("1.9") === 0; // Gecko 1.9.2
  },

  _getTopic: function(is192) {
    return is192 ? "content-document-global-created" : "document-element-inserted";
  },

  QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),
  observe: function(subject, topic, data) {
    var win;
    var requestUri;

    switch (topic) {
      case "content-document-global-created": // Gecko 1.9.2
        win = subject;
        if (win === null) {
          return;
        }
        if (data === "null") {
          return;
        }
        var io = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
        requestUri = io.newURI(data, null, null);
        break;
      case "document-element-inserted":
        win = subject.defaultView;
        if (win === null) {
          return; // xsl/xbl
        }
        requestUri = subject.documentURIObject;
        break;
      default:
        throw new Error(topic);
    }

    if (win.wrappedJSObject.hasOwnProperty("_multifox_obs")) {
      console.warn("DUP " + topic + " @ " + requestUri.spec); // moz bug?
      return;
    }
    win.wrappedJSObject._multifox_obs = true; // BUG visible by content

    // requestUri=about:neterror?e=dnsNotFound&u=http...
    var tabUser = TabLogin.getFromWindow(win);
    if (tabUser.tabElement === null) {
      return;
    }

    var tab = tabUser.tabElement;

    if (TabLogin.isLoginInProgress(tab)) {
      tabUser = TabLogin.getLoginInProgress(tab);
    }

    if (isTopWindow(win) === false) {
      // iframe
      if (isSupportedScheme(requestUri.scheme) && shouldSubElementUseLogin(tabUser, requestUri)) {
        this._inject(win);
      }
      return;
    }

    if (isSupportedScheme(requestUri.scheme)) {
      var usr = tabUser.loginUser16;
      if ((usr !== null) && (usr !== TabLogin.NewAccount)) {
        this._inject(win);
      }
    } else { // about, javascript
      tab.removeAttribute("multifox-tab-current-tld");
      TabLogin.removeLogin(tab);
    }

    updateUI(tab);
    RedirDetector.resetTab(tab);
  },


  _inject: function(win) {
    var sandbox = Components.utils.Sandbox(win);
    sandbox.window = win.wrappedJSObject;
    sandbox.document = win.document.wrappedJSObject;

    var src = this._loader.getScript();
    try {
      Components.utils.evalInSandbox(src, sandbox);
    } catch (ex) {
      showError(win, "sandbox", win.document.location + " " + "//exception=" + ex);
    }
  }
};


function isSupportedScheme(scheme) {
  return (scheme === "http") || (scheme === "https");
}


function throwException() {
  console.trace();
  throw new Error("throwException");
}


function ScriptSourceLoader(is192) {
  if (is192) {
    // Gecko 1.9.2
    this._path = "chrome://multifox/content/content-injection-192.js";
  } else {
    // Gecko 2.0
    this._path = "chrome://multifox/content/content-injection.js";
  }
  this._src = null;
  this._load(true);
}


ScriptSourceLoader.prototype = {
  getScript: function() {
    if (this._src === null) {
      this._load(false);
    }
    return this._src;
  },

  _load: function(async) {
    var me = this;
    var xhr = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Ci.nsIXMLHttpRequest);
    xhr.onload = function() {
      delete me._path;
      me._src = xhr.responseText + "initContext(window, document, '"
                                 + m_runner.eventSentByChrome + "','" + m_runner.eventSentByContent + "');";
    };
    xhr.open("GET", this._path, async);
    xhr.overrideMimeType("text/plain");
    xhr.send(null);
  }
};
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const httpListeners = {
  request: {
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),

    // nsIObserver
    observe: function REQUEST(aSubject, aTopic, aData) {

try{

      var httpChannel = aSubject.QueryInterface(Ci.nsIHttpChannel);
      var win = getChannelWindow(httpChannel);
      if (win === null) {
        return; // from component
      }

      if (isDocumentChannel(httpChannel)) {
        console.log("===>request", win === win.top ? "top" : "frame", httpChannel.requestMethod, httpChannel.URI.spec, "@", win.document.href);
      }

      var tabLogin = getRequestChannelUser(httpChannel, win);
      if ((tabLogin === null) || (tabLogin.loginTld === null)) {
        return; // use default session
      }

      var myHeaders = HttpHeaders.fromRequest(httpChannel);
      if (myHeaders["authorization"] !== null) {
        showError(win, "authorization", "-");
        return;
      }

      var cook = Cookies.getCookie(false, httpChannel.URI, tabLogin);
      httpChannel.setRequestHeader("Cookie", cook, false);



}catch(ex){
  console.trace();
  Components.utils.reportError(ex);
}


    }
  },

  response: {
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),

    // nsIObserver
    observe: function RESPONSE(aSubject, aTopic, aData) {
      var httpChannel = aSubject.QueryInterface(Ci.nsIHttpChannel);
      var win = getChannelWindow(httpChannel);
      if (win === null) {
        return;
      }

      var tabLogin = getResponseTabUser(httpChannel, win);
      if ((tabLogin === null) || (tabLogin.loginTld === null)) {
        // use original Set-Cookie
        return;
      }

      var myHeaders = HttpHeaders.fromResponse(httpChannel);
      if (myHeaders["www-authenticate"] !== null) {
        showError(win, "www-authenticate", "-");
        return;
      }

      var setCookies = myHeaders["set-cookie"];
      if (setCookies === null) {
        return;
      }

      // replace "Set-Cookie"
      httpChannel.setResponseHeader("Set-Cookie", null, false);
      Cookies.setCookie(tabLogin, httpChannel.URI, setCookies, false);
    }
  }
};


// pageshow event => call updateUI for non http/https protocols and cached pages
function updateNonNetworkDocuments(evt) {
  var doc = evt.target;
  var uri = doc.documentURIObject;
  if (isTopWindow(doc.defaultView) === false) {
    return;
  }

  if (isSupportedScheme(uri.scheme)) {
    // BUG rightclick>show image ==> evt.persisted=false
    if (evt.persisted) {
      // http top doc from cache
      var tabLogin = TabLogin.getFromWindow(doc.defaultView);
      var tab = tabLogin.tabElement;
      if (tab !== null) {
        RedirDetector.invalidateTab(tab); // invalidate redir before calling defineTopDocLogin
        var dummy = defineTopDocLogin(uri, tabLogin);
        updateUI(tab);
      }
    }

  } else { // about: etc
    // request listener was not called
    var tab = TabLogin.getFromWindow(doc.defaultView).tabElement;
    if (tab) {
      tab.removeAttribute("multifox-tab-current-tld");
      RedirDetector.resetTab(tab);
      TabLogin.removeLogin(tab);
      updateUI(tab);
    }
  }

}


const HttpHeaders = {
  visitLoop: {
    values: null,
    visitHeader: function(name, value) {
      var n = name.toLowerCase();
      if (n in this.values) {
        this.values[n] = value;
      }
    }
  },

  fromRequest: function(request) {
    var nameValues = {
      //"cookie": null, //for debug only
      "authorization": null
    }
    this.visitLoop.values = nameValues;
    request.visitRequestHeaders(this.visitLoop);
    return nameValues;
  },

  fromResponse: function(response) {
    var nameValues = {
      "set-cookie": null,
      "www-authenticate": null
    }
    this.visitLoop.values = nameValues;
    response.visitResponseHeaders(this.visitLoop);
    return nameValues;
  }
};


function getChannelWindow(channel) {
  if (channel.notificationCallbacks) {
    try {
      return channel
              .notificationCallbacks
              .getInterface(Ci.nsILoadContext)
              .associatedWindow;
    } catch (ex) {
      //util2.logEx("channel.notificationCallbacks ", channel.notificationCallbacks, channel.URI.spec, ex);
    }
  }

  if (channel.loadGroup && channel.loadGroup.notificationCallbacks) {
    try {
      return channel
              .loadGroup
              .notificationCallbacks
              .getInterface(Ci.nsILoadContext)
              .associatedWindow;
    } catch (ex) {
      util2.logEx("channel.loadGroup", channel.loadGroup, channel.URI.spec, ex);
    }
  }

  //var isChrome = context.associatedWindow instanceof Ci.nsIDOMChromeWindow;
  //return context.isContent ? context.associatedWindow : null;
  //console.log("LOAD CONTEXT FAIL " + channel.URI.spec);
  return null; // e.g. <link rel=prefetch> <link rel=next> ...
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


/*

BROWSING PROTOCOL


UPDATING TAB ATTRIBUTES: "multifox-tab-login-tld" + "multifox-tab-login-user16"
========================================================================

*** START => SUBMIT
- in progress => user/pw submitted and copied to tab attr: "multifox-login-submit-domain" + "multifox-login-submit-user"

*** END => submitLandPageFound
- remove attrs multifox-login-submit-*
- success => commit: setLogin
- fail => restore data domain

*** REQUEST
- login in progress?
  - channel: LOAD_DOCUMENT_URI + top document (ignore subdocs)
  - POST request detected ==> moveData_loginSubmitted
  - move cookies/data to a namespace


*** RESPONSE



*** DOCUMENT CREATED
*/


function getRequestChannelUser(channel, win) {
  var tabLogin = TabLogin.getFromWindow(win);

  if (tabLogin.tabElement === null) {
    var chromeWin = WindowParents.getChromeWindow(win);
    if (chromeWin && (chromeWin.document.documentURI === "chrome://global/content/viewSource.xul")) {
      // view source window
      return OpenerWindowIdentity.findFromOpenerSelectedTab(win);
    }
    console.log("getRequestChannelUser", "NO TAB", channel.URI.spec);
    return null; // request from chrome (favicon, updates)
  }


  var tab = tabLogin.tabElement;
  var requestUri = channel.URI;


  if (isSupportedScheme(requestUri.scheme) === false) {
    // tabLogin is invalid ==> use default session
    console.log('getRequestChannelUser', requestUri.scheme);
    tab.removeAttribute("multifox-tab-current-tld");
    TabLogin.removeLogin(tab);
    return null;
  }

  if (isDocumentChannel(channel) === false) {
    // img, css...
    if (TabLogin.isLoginInProgress(tab)) {
      tabLogin = TabLogin.getLoginInProgress(tab);
    }
    return shouldSubElementUseLogin(tabLogin, requestUri) ? tabLogin : null;
  }

  // channel will create a document

  if (TabLogin.isLoginInProgress(tab)) {
    // obs: getLoginInProgress and getFromWindow may both be valid (and different)
    return moveData_loginSubmitted(tab); // login may be successful, so move cookies now. it may be canceled later.
  }

  if (isTopWindow(win)) {
    // channel will replace the current top document
    return defineTopDocLogin(requestUri, tabLogin);
  } else {
    // iframes shouldn't modify tab identity (obs: login iframes are allowed)
    return shouldSubElementUseLogin(tabLogin, requestUri) ? tabLogin : null;
  }
}


function defineTopDocLogin(requestUri, prevLogin) {
  var tab = prevLogin.tabElement;
  var redirInvalidated = tab.hasAttribute("multifox-redir-invalidate");
  if (redirInvalidated) {
    RedirDetector.resetTab(tab);
  }

  var requestTld = m_tld.getBaseDomain(requestUri);
  var prevTopTld = null; // null => new tab or unsupported scheme
  if (tab.hasAttribute("multifox-tab-current-tld")) {
    prevTopTld = tab.getAttribute("multifox-tab-current-tld");
    if (prevTopTld === requestTld) {
      // prevLogin is still valid!
      // BUG login happens in another tab => this one will keep anonymous (LoginData.getDefaultLogin is never called)
      return prevLogin.loginUser16 === TabLogin.NewAccount ? null : prevLogin;
    }
  }

  // New TLD! (prevLogin is [probably] invalid for the new top document)

  var newLogin = defineNewTldLogin(tab, requestTld, prevLogin, prevTopTld, redirInvalidated);
  if (newLogin !== null) {
    TabLogin.setLogin(newLogin);
  } else {
    TabLogin.removeLogin(tab); // default session
  }
  return newLogin;
}


function defineNewTldLogin(tab, requestTld, prevLogin, prevTopTld, redirInvalidated) {
  var prevPrevTopTld = tab.hasAttribute("multifox-tab-previous-tld")
                     ? tab.getAttribute("multifox-tab-previous-tld") : null;

  tab.setAttribute("multifox-tab-previous-tld", prevTopTld);
  tab.setAttribute("multifox-tab-current-tld", requestTld);

  // we know 3 TLDs:
  //   requestTld     - new tld
  //   prevTopTld     - previous tld (associated with prevLogin)
  //   prevPrevTopTld - previous to previous tld

  console.log('new tld /', prevPrevTopTld, "==>", prevTopTld, '==>', requestTld);

  // cross login in progress
  var commitCrossLogin = inheritFromPreviousTld(redirInvalidated, prevPrevTopTld, prevTopTld, requestTld, prevLogin);
  if (commitCrossLogin) {
    console.log("commitCrossLogin", redirInvalidated, prevPrevTopTld, prevTopTld, requestTld, prevLogin);
    //tab.ownerDocument.defaultView.alert("commitCrossLogin "+ requestTld);
    moveData_fromDefault(requestTld, prevLogin);
    tab.setAttribute("multifox-cross-login-commited-wait-landing-page", "true");
    tab.linkedBrowser.addEventListener("DOMContentLoaded", checkEmptyPage, false);
    return prevLogin;
  }

  // cross login done
  if (tab.hasAttribute("multifox-cross-login-commited-wait-landing-page")) {
    return prevLogin;
  }

  // popups
  if (tab.ownerDocument.documentElement.hasAttribute("multifox-window-uninitialized")) {
    tab.ownerDocument.documentElement.removeAttribute("multifox-window-uninitialized");
    prevLogin = OpenerWindowIdentity.findFromWindow(tab.linkedBrowser.contentWindow);
    prevLogin.tabElement = tab;
  }

  // may keep user => different tld, same login
  return LoginData.getDefaultLogin(requestTld, prevLogin);
}


function getResponseTabUser(channel, win) {
  var tabLogin = TabLogin.getFromWindow(win);
  if (tabLogin.tabElement === null) {
    return null; // request from chrome (favicon, updates)
  }

  if (TabLogin.isLoginInProgress(tabLogin.tabElement)) {
    tabLogin = TabLogin.getLoginInProgress(tabLogin.tabElement);
  } else {
    if (tabLogin.loginTld === null) {
      return null;
    }
  }

  if (isDocumentChannel(channel) && isTopWindow(win)) {
    // channel will replace window
    return tabLogin;
  }
  return shouldSubElementUseLogin(tabLogin, channel.URI) ? tabLogin : null;
}


// tab element = css, img, iframe, ... ==> not called for top doc, so tabLogin is already valid
function shouldSubElementUseLogin(tabLogin, elemUri) {
  var usr = tabLogin.loginUser16;
  if ((usr === null) || (usr === TabLogin.NewAccount)) {
    return false;
  }

  // tldLogin = google.com
  // tab host = www.youtube.com
  // elemUri  = img.youtube.com
  var tldLogin = tabLogin.loginTld;
  var tldElem = m_tld.getBaseDomain(elemUri);
  if (tldElem === tldLogin) {
    return tabLogin;
  }

  // TODO use "multifox-tab-current-tld" instead
  //var tabTld = m_tld.getBaseDomain(tabLogin.tabElement.linkedBrowser.currentURI);
  console.assert(tabLogin.tabElement.hasAttribute("multifox-tab-current-tld"), "attr multifox-tab-current-tld not found");
  var tabTld = tabLogin.tabElement.getAttribute("multifox-tab-current-tld");
  return tldElem === tabTld ? true : LoginData.hasLogin(tldElem, tabLogin);
}





/*
CROSS-TLD LOGIN PROTOCOL
TLD COPY LOGIN PROTOCOL
========================
*/

function inheritFromPreviousTld(redirInvalidated, prevPrevTopTld, prevTopTld, requestTld, prevLogin) {
  if (redirInvalidated ||
     (prevPrevTopTld === null) || (prevTopTld === null) ||
     (prevLogin === null)      || (prevLogin.loginTld === null)) {
    return false;
  }

  console.log('REDIR! 0');
  return (prevPrevTopTld === requestTld) &&
         (LoginData.hasTld(requestTld) === false) &&
         LoginData.hasTld(prevTopTld);
}


function checkEmptyPage(evt) {
  var contentDoc = evt.target;
  console.log('checkEmptyPage top=', contentDoc.defaultView === contentDoc.defaultView.top, contentDoc.location);
  if (isEmptyPage(contentDoc)) {
    return;
  }
  var tabLogin = TabLogin.getFromWindow(contentDoc.defaultView);
  var tab = tabLogin.tabElement;
  tab.removeAttribute("multifox-cross-login-commited-wait-landing-page");
  tab.linkedBrowser.removeEventListener("DOMContentLoaded", checkEmptyPage, false);
}


// ugly workaround, but it works most of the time
var RedirDetector = {
  onMouseDown: function(evt) {
    if (evt.button === 0) {
      RedirDetector._invalidate(evt.target.ownerDocument.defaultView);
    }
  },

  onKey: function(evt) {
    // keypress--> charCode=0
    //console.log('onkey',evt.altKey, evt.ctrlKey,evt.keyCode,evt.charCode,evt.keyCode !== 0  );
    var k = Ci.nsIDOMKeyEvent;
    //if (evt.altKey || evt.ctrlKey || (evt.keyCode !== 0) || (evt.charCode === k.DOM_VK_SPACE)) {
    if (evt.keyCode < 33) {
      // keyCode ==> backspace, enter
      // DOM_VK_SPACE ==> submit form
      RedirDetector._invalidate(evt.target.ownerDocument.defaultView);
    }
  },

  _invalidate: function(win) {
    if (win === null) {
      return;
    }
    var tab = TabLogin.getFromWindow(win).tabElement;
    if (tab === null) {
      if ("getBrowser" in win) {
        tab = win.getBrowser().selectedTab;
      }
    }
    RedirDetector.invalidateTab(tab);
  },

  invalidateTab: function(tab) {
    if ((tab === null) || tab.hasAttribute("multifox-redir-invalidate")) {
      return;
    }

    tab.setAttribute("multifox-redir-invalidate", "true");
  },

  resetTab: function(tab) {
    tab.removeAttribute("multifox-redir-invalidate");
  }
};


function isDocumentChannel(channel) {
  return (channel.loadFlags & Ci.nsIChannel.LOAD_DOCUMENT_URI) !== 0;
}


function isTopWindow(win) {
  return win === win.top;
}


function isEmptyPage(contentDoc) {
  //var a = contentDoc.getElementsByTagName("head")[0].getElementsByTagName("*").length;
  var body = contentDoc.getElementsByTagName("body");
  var tags = body.length > 0 ? body[0].getElementsByTagName("*").length : 0;

  // TODO
  /*
  var all = contentDoc.forms;
  var qty = 0;
  for (var idx = all.length - 1; idx > -1; idx--) {
    qty += countPasswords(all[idx]);
  }
  */
  console.log("isEmptyPage", tags < 15, contentDoc.location);
  return tags < 15;
}





/*
FORM LOGIN
  SUBMIT
    Land Page
      Fail
      Success




browse ANON_TLD
[...]

browser AUTH_TLD
save multifox-cross-login-try-anon-tld=ANON_TLD
[...]

returns to ANON_TLD. Should add ANON_TLD to AUTH_TLD space?
- redir? yes!
- just made a login in AUTH_TLD? yes!
- How?
  - click on last page?
  - redir

- redir se nao houve click/key/fav na last page setAtribute("multifox-last-page-had-user-action", "") (nao contar chrome ev)



ANON_TLD_2 (logged or not)
  - set watched-tld="new_tld_1"
  - land page:
    - login form => abort() (use form protocol)
  - redirect to "new_tld_1"

ANON_TLD_1
*/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


/*
LOGIN
=====
  post + login form: submit detected
    save username
      request will isolate data to login tld
      wait land page (it can be in another tld)
        success: commit
        fail: undo move data
*/

var SubmitObserver = {
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIFormSubmitObserver]),

  notify: function (form, win, actionURI, cancelSubmit) {
    console.assert(form.ownerDocument.defaultView === win, "form.ownerDocument.defaultView != win");
    this._notify(form, win);
    return true;
  },

  _notify: function(form, win) {
    if (form.method.toUpperCase() !== "POST") {
      return;
    }

    var uri = win.document.documentURIObject;
    if (isSupportedScheme(uri.scheme) === false) {
      return;
    }

    var tabLogin = TabLogin.getFromWindow(win);
    var tab = tabLogin.tabElement;
    if (tab === null) {
      return;
    }

    if (countPasswords(form, true) === 0) {
      return; // no password field => not a login form
    }

    var username = findUserName(form);
    if (username === null) {
      if (tabLogin.loginTld !== null) {
        console.log("SubmitObserver: NOP (username not found, confirm pw form?)");
        return;
      }
      console.log("SubmitObserver: NOP (username not found)");
      return; // TODO username = "random" or error message (icon);
    }
    console.assert(username.length > 0, "SubmitObserver: findUserName");

    // keep tab identity, request will need to copy its cookies
    TabLogin.setLoginInProgress(tab, username, m_tld.getBaseDomain(uri));
    tab.linkedBrowser.addEventListener("DOMContentLoaded", SubmitObserver._detectSubmitLandPage, false);
  },


  _detectSubmitLandPage: function(evt) {
    var contentDoc = evt.target;
    if (isEmptyPage(contentDoc)) {
      return;
    }

    // land page found!
    var tab = TabLogin.getFromWindow(contentDoc.defaultView).tabElement;
    tab.linkedBrowser.removeEventListener("DOMContentLoaded", SubmitObserver._detectSubmitLandPage, false);


    var newLogin = TabLogin.getLoginInProgress(tab);
    var moveMode = TabLogin.removeLoginInProgress(tab);
    console.assert(newLogin.loginTld    !== null, "multifox-login-submit-domain");
    console.assert(newLogin.loginUser16 !== null, "multifox-login-submit-user");


    if (hasLoginForm(contentDoc)) {
      console.log("submitLandPageFound", "login error detected", contentDoc.documentURI);
      // cancel login
      if (moveMode === "user") {
        moveMultifoxDomainDataToAnotherUser_cancel_login(newLogin);
      } else {
        SubmitObserver._moveDomainDataToDefault_cancel_login(newLogin);
      }

    } else {
      // commit login!
      console.log("submitLandPageFound", "login successful detected", contentDoc.documentURI);
      TabLogin.setLogin(newLogin);
      updateUI(tab, true);
    }
  },

  // TODO moveDomainDataTo ... ***previous domain***
  _moveDomainDataToDefault_cancel_login: function(tabLogin) { // restoreDataDomain
    var srcHost = toInternalHost(tabLogin.loginTld, tabLogin);
    var all = removeTldData_cookies(srcHost);

    console.log('===>_moveDomainDataToDefault_cancel_login ', tabLogin.loginTld, "-", tabLogin.loginUser16, "=", srcHost, "_", all.length);

    for (var idx = all.length - 1; idx > -1; idx--) {
      var cookie = all[idx];
      var hostData = getLoginFromInternalHost(cookie.host);
      if (hostData !== null) {
        copyCookieToNewHost(cookie, hostData.realHost);
      }
    }

    var all2 = removeTldData_LS(srcHost);

  }

};


function moveMultifoxDomainDataToAnotherUser_cancel_login(tabLogin) { // TODO
}



function moveDataToAnotherUser(tabTld, newLogin, prevLogin) {
  console.assert(prevLogin.loginTld === newLogin.loginTld, "moveDataToAnotherUser tld");
  if (prevLogin.loginUser16 === newLogin.loginUser16) {
    if (prevLogin.loginTld === newLogin.loginTld) {
      return;
    }
  }

  var tld = toInternalHost(tabTld, prevLogin);
  var all = removeTldData_cookies(tld);
  console.log("moveDataToAnotherUser", tabTld, all.length, "cookies.", prevLogin.loginUser16, newLogin.loginUser16);
  var cookie;
  var hostData;
  for (var idx = all.length - 1; idx > -1; idx--) {
    cookie = all[idx];
    hostData = getLoginFromInternalHost(cookie.host);
    if (hostData !== null) {
      copyCookieToNewHost(cookie, toInternalHost(hostData.realHost, newLogin));
    }
  }

  var all2 = removeTldData_LS(tld);
}


// isolate cookies from domain
function moveData_fromDefault(domain, tabLogin) { // BUG se tabLogin.loginUser16="" vai pensar NewAccount
  var all = removeTldData_cookies(domain);
  console.log('===>moveData_fromDefault ', domain, "loginTld", tabLogin.loginTld, "loginUser:", tabLogin.loginUser16, "cookies:", all.length);
  var cookie;
  for (var idx = all.length - 1; idx > -1; idx--) {
    cookie = all[idx];
    copyCookieToNewHost(cookie, toInternalHost(cookie.host, tabLogin));
  }

  var all2 = removeTldData_LS(domain);
}


function hasLoginForm(doc) { // TODO check frames
  if ("forms" in doc === false) {
    return false; // xml
  }

  var all = doc.forms;
  for (var idx = all.length - 1; idx > -1; idx--) {
    var qty = countPasswords(all[idx], false);
    if (qty === 1) {
      return true;
    }
  }

  return false;
}


function countPasswords(form, checkEmpty) {
  var qty = 0;
  var all = form.elements;
  var INPUT = Ci.nsIDOMHTMLInputElement;
  for (var idx = all.length - 1; idx > -1; idx--) {
    var elem = all[idx];
    if ((elem instanceof INPUT) && (elem.type === "password")) {
      if ((checkEmpty === false) || isElementVisibleOrNotEmpty(elem)) {
        qty++;
      }
    }
  }
  return qty;
}


function findUserName(form) {
  var INPUT = Ci.nsIDOMHTMLInputElement;
  var lastTextField = null;
  var all = form.elements;
  var elem;

  for (var idx = 0, len = all.length; idx < len; idx++) {
    elem = all[idx];
    if ((elem instanceof INPUT) === false) {
      continue;
    }
    switch (elem.type) {
      case "text":
      case "email":
      case "url":
      case "tel":
      case "number":
      case "password":
        break;
      default:
        continue;
    }
    if (isElementVisibleOrNotEmpty(elem) === false) {
      // ignore empty/hidden fields
      console.log("element ignored", elem.name);
      continue;
    }
    if (elem.type === "password") {
      if (lastTextField !== null) {
        return lastTextField;
      }
    } else {
      lastTextField = elem.value;
    }
  }
  return lastTextField;
}


function isElementVisibleOrNotEmpty(elem) {
  // display:none/visibility:hidden elements are removed from form anyway
  //console.log(elem, elem.tagName, elem.getBoundingClientRect);
  console.assert(elem.getBoundingClientRect().width > 0, elem.tagName + " name-" + elem.name + "=" + elem.value + " getBoundingClientRect().width" + elem.getBoundingClientRect().width);

  var win = elem.ownerDocument.defaultView;
  var val = win.getComputedStyle(elem, "").getPropertyValue("visibility");
  console.assert(val === "visible", "visibility", val);

  return elem.value.trim().length > 0;
}


function moveData_loginSubmitted(tab) {
  var newLogin = TabLogin.getLoginInProgress(tab);
  if (TabLogin.hasLoginInProgressMoveData(tab)) {
    return newLogin;
  }

  console.log("moveData_loginSubmitted");
  var currentLogin = TabLogin.getFromTab(tab);

  // anonymous tab
  if (currentLogin.loginUser16 === null) {
    console.log("moveData_fromDefault logged=anon");
    TabLogin.setLoginInProgressMoveData(tab, "default");
    moveData_fromDefault(newLogin.loginTld, newLogin);

  // new account tab
  } else if (currentLogin.loginUser16 === TabLogin.NewAccount) {
    console.log("moveData_fromDefault logged=NewAccount");
    TabLogin.setLoginInProgressMoveData(tab, "default");
    TabLogin.removeLogin(tab);
    moveData_fromDefault(newLogin.loginTld, newLogin);

  // new login from a logged in tab
  } else {
    // tab is marked as logged in, but it is actually logged out
    TabLogin.setLoginInProgressMoveData(tab, "user");
    TabLogin.removeLogin(tab);
    console.log("moveDataToAnotherUser logged=", currentLogin.loginUser16, newLogin.loginTld, newLogin.loginUser16);
    moveDataToAnotherUser(newLogin.loginTld, newLogin, currentLogin);
  }

  return newLogin;
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


var LoginData = {
  _forms: null,
  _sites: null,

  _reset: function() {
    this._forms = { __proto__: null };
    this._sites = { __proto__: null };
  },

  shutdown: function() {
    this._invalidate();
  },

  _invalidate: function() {
    this._forms = null;
    this._sites = null;
  },

  _ensureValid: function() {
    if (this._sites === null) {
      this._update();
    }
  },

  /*
  this._cache = {
    "orkut.com": [
      "efghijklmnop.google.com",
      "abcdefghijkl.google.com"
    ],

    "google.com": [
      "efghijklmnop.google.com",
    ],

    "google.com.br": [
      "efghiop.google.com",
      "abcdefghijkl.google.com"
    ]};



  this._forms = {
    "google.com": [
      "efghijklmnop",
      "abcdefghijkl"
    ],

  this._sites = {
    "orkut.com": [
      "google.com"
    ],
  */
  _update: function() {
    console.log("LoginData._update");
    this._reset(); // BUG default logins are lost!
    var hostData;
    var users;
    var host;
    var all = getAllCookiesFromHost("x-content.x-namespace");

    for (var idx = all.length - 1; idx > -1; idx--) {
      hostData = getLoginFromInternalHost(all[idx].host);
      if (hostData === null) {
        continue;
      }
      host = hostData.realHost;
      var idx2 = host.indexOf(".");
      if (idx2 === 0) { // ".google.com"
        host = host.substr(1);
        var idx3 = host.indexOf(".");
        if (idx3 > -1) { // "localhost" => tld exception
          host = m_tld.getBaseDomainFromHost(host);
        }
      } else if (idx2 > 0) { // idx2=-1 >>> "localhost" >>> tld exception
        host = m_tld.getBaseDomainFromHost(host);
      }

      /*
      this._forms = {
        "google.com": [
          "efghijklmnop",
          "abcdefghijkl"
        ],
        "twtter.com": [
          "efghijklmnop",
        ]

      this._sites = {
        "my-app.com": [
          "google.com",
          "facebook.com",
          "twitter.com"
        ],
        "orkut.com": [
          "google.com"
        ],
      */

      var login2 = splitUserNameAndTld(hostData.fullLogin);
      var formUsr = login2[0];
      var formTld = login2[1];

      if (formTld in this._forms) {
        var aa = this._forms[formTld];
        if (aa.indexOf(formUsr) === -1) {
          aa.push(formUsr);
        }
      } else {
        this._forms[formTld] = [formUsr];
      }


      if (host in this._sites) {
        var bb = this._sites[host];
        if (bb.indexOf(formTld) === -1) {
          bb.push(formTld);
        }
      } else {
        this._sites[host] = [formTld];
      }

    }

    console.log('_update\n' + JSON.stringify(this._sites, null, 2) + "\n==========\n" + JSON.stringify(this._forms, null, 2));
  },


  hasLogin: function(tabTld, tabLogin) {
    var login = tabLogin.loginUser16 + "." + tabLogin.loginTld;
    return this.hasLogin_2(tabTld, login);
    /*
    if (this.hasTld(tabTld)) {
      var user = tabLogin.loginUser16 + "." + tabLogin.loginTld;
      return this._cache[tabTld].indexOf(user) > -1;
    }
    return false;
    */
  },


  hasLogin_2: function(tabTld, login) {
    // deve retornar true se existir o mesmo username p/ user.loginTld
    this._ensureValid();
    if (tabTld in this._sites) {
      var dd = this._sites[tabTld];
      var login2 = splitUserNameAndTld(login);
      var formUsr = login2[0];
      var formTld = login2[1];
      if (dd.indexOf(formTld) > -1) {
        if (formTld in this._forms) {
          var uu = this._forms[formTld];
          if (uu.indexOf(formUsr) > -1) {
            return true;
          }
        }
      }
    }
    return false;
  },

  hasTld: function(tabTld) { // hasSite isLoggedTld
    this._ensureValid();
    return tabTld in this._sites;
  },

  getTldUsers: function(tabTld) { // tabTld = "orkut.com.br" (login: "XYZ.google.com")
    this._ensureValid();
    if ((tabTld in this._sites) === false) {
      return [];
    }

    // TODO test multiple sites eg bugzilla + amo
    var tabUsers = [];
    var tlds = this._sites[tabTld];
    for (var idx1 = tlds.length - 1; idx1 > -1; idx1--) {
      var loginTld = tlds[idx1];
      var users = this._forms[loginTld];
      for (var idx2 = users.length - 1; idx2 > -1; idx2--) {
        var myUser = users[idx2];
        tabUsers.push({ user16:myUser, username:hexToString(myUser), tld:loginTld });
      }
    }

    // alphabetical sort
    tabUsers.sort(function(userA, userB) {
      return userB.username.localeCompare(userA.username);
    });

    return tabUsers;
  },


  setTabAsDefaultLogin: function(tab) {
    if (tab.hasAttribute("multifox-tab-current-tld") === false) {
      console.trace("LoginData.setTabAsDefaultLogin -- no attr multifox-tab-current-tld");
      return;
    }

    var tabLogin = TabLogin.getFromTab(tab);
    if (tabLogin.loginUser16 === TabLogin.NewAccount) {
      return;
    }

    var tabTld = tab.getAttribute("multifox-tab-current-tld");
    this.setDefaultLogin(tabTld, tabLogin.loginUser16, tabLogin.loginTld);
  },


  setDefaultLogin: function(tabTld, user, tld) {
    this._ensureValid();
    console.log("LoginData.setDefaultLogin", user, tld);

    if ((tabTld in this._sites) === false) {
      return;
    }

    if ((tld in this._forms) === false) {
      return;
    }

    var allTlds = this._sites[tabTld];
    var idx2 = allTlds.indexOf(tld);
    if (idx2 > 0) {
      allTlds.splice(idx, 1);
      allTlds.unshift(tabTld); // [0] => default
      console.log("LoginData.setTabAsDefaultLogin2 -- updated");
    }

    var allUsers = this._forms[tld];
    if (user === TabLogin.NewAccount) {
      allUsers.unshift(TabLogin.NewAccount); // workaround (getDefaultLogin will remove it)
    } else {
      var idx = allUsers.indexOf(user);
      // idx = -1 ==> login not found (cookies cleared?)
      // idx =  0 ==> user is already the default
      if (idx > 0) {
        allUsers.splice(idx, 1);
        allUsers.unshift(user); // [0] => default
        console.log("LoginData.setTabAsDefaultLogin -- updated", hexToString(user), user, tld, this._forms[tld].toSource());
        //console.log("LoginData.setTabAsDefaultLogin -- updated" + , hexToString(user), user, tld, this._forms[tld].toSource());
        //console.assert(this._forms[tld][0] === user, "this._forms[tld];");
      }
    }
  },


  getDefaultLogin: function(tabTld, tabLogin) {
    this._ensureValid();

    if ((tabTld in this._sites) === false) {
      return null;
    }

    // obs: loginUser16/loginTld may be null
    var loginTlds = this._sites[tabTld];
    var defaultFormTld = loginTlds[0];
    if (defaultFormTld === tabLogin.loginTld) {
      console.log("LoginData.getDefaultLogin", "don't change user", hexToString(tabLogin.loginUser16));
      return tabLogin;
    }

    console.assert(defaultFormTld in this._forms, "this._forms");
    var users = this._forms[defaultFormTld];
    console.assert(users.length > 0, "users.length");
    var defaultUser = users[0];

    var newLogin = {
      tabElement:  tabLogin.tabElement,
      loginUser16: defaultUser,
      loginTld:    defaultFormTld
    };

    if (defaultUser === TabLogin.NewAccount) {
      users.splice(0, 1); // workaround - middle click on "New Account" -- BUG next calls won't be NewAccount
      console.log("LoginData.getDefaultLogin", "user NEW ", defaultFormTld, tabLogin.loginTld, tabLogin.loginUser16);
    }

    console.log("LoginData.getDefaultLogin", "user found ", defaultFormTld, defaultUser, hexToString(defaultUser));
    return newLogin;

  },


  onCookieRejected: {
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),
    observe: function(subject, topic, data) {
      console.log("cookie-rejected:");
      console.log("cookie-rejected\n", subject, "\n", topic, "\n", data, "\n", subject.QueryInterface(Ci.nsIURI).spec);
    }
  },

  onCookieChanged: {
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),
    observe: function(subject, topic, data) {
      if (LoginData._sites === null) {
        return;
      }
      switch (data) {
        case "changed":
          break;
        case "added":
          var cookie = subject.QueryInterface(Ci.nsICookie2);
          if (hasRootDomain("x-content.x-namespace", cookie.host)) {
            var hostData = getLoginFromInternalHost(cookie.host);
            if ((hostData !== null) && LoginData.hasLogin_2(hostData.realHost, hostData.fullLogin) === false) {
              LoginData._invalidate(); // login added
            }
          }
          break;
        case "deleted":
          var cookie = subject.QueryInterface(Ci.nsICookie2);
          if (hasRootDomain("x-content.x-namespace", cookie.host)) {
            LoginData._invalidate(); // login removed?
          }
          break;
        case "batch-deleted":
          var all = subject.QueryInterface(Ci.nsIArray).enumerate();
          console.log("cookieChanged!!!!!! " + data + all);
          while (all.hasMoreElements()) {
            var cookie = all.getNext().QueryInterface(Ci.nsICookie2);
            console.log("cookieChanged batch-deleted>> " + cookie.host);
            if (hasRootDomain("x-content.x-namespace", cookie.host)) {
              LoginData._invalidate(); // login removed?
              break; // it is not necessary yo continue
            }
          }
          break;
        case "cleared":
          LoginData._reset();
          break;
        case "reload":
          LoginData._invalidate();
          break;
        default:
          LoginData._invalidate(); // ???
          break;
      }
    }
  }
};


function splitUserNameAndTld(usrTld) {
  var idx = usrTld.indexOf(".");
  return [usrTld.substr(0, idx),
          usrTld.substr(idx + 1)];
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


function documentCookie(obj, contentDoc) {
  switch (obj.cmd) {
    case "set":
      documentCookieSetter(obj, contentDoc);
      return undefined;
    case "get":
      return documentCookieGetter(obj, contentDoc);
    default:
      throw new Error("documentCookie " + obj.cmd);
  }
}


function documentCookieGetter(obj, contentDoc) {
  var tabLogin = TabLogin.getFromWindow(contentDoc.defaultView);
  if (tabLogin.tabElement === null) {
    return;
  }

  if (TabLogin.isLoginInProgress(tabLogin.tabElement)) {
    tabLogin = TabLogin.getLoginInProgress(tabLogin.tabElement);
  }

  console.assert(tabLogin.loginTld !== null, "documentCookieGetter not logged=" + tabLogin.loginUser16 + tabLogin.loginTld + "," + contentDoc.documentURI);
  var cookie2 = Cookies.getCookie(true, contentDoc.documentURIObject, tabLogin);

  var cookie = cookie2 === null ? "" : cookie2;
  return cookie; // send cookie value to content
}


function documentCookieSetter(obj, contentDoc) {
  var tabLogin = TabLogin.getFromWindow(contentDoc.defaultView);
  if (tabLogin.tabElement === null) {
    return;
  }

  if (TabLogin.isLoginInProgress(tabLogin.tabElement)) {
    //console.log('documentCookieSetter login' + tabLogin.loginTld);
    tabLogin = TabLogin.getLoginInProgress(tabLogin.tabElement);
    //console.log('documentCookieSetter login' + tabLogin.loginTld);
  }

  console.assert(tabLogin.loginTld !== null, "documentCookieSetter not logged=" + tabLogin.loginTld);
  Cookies.setCookie(tabLogin, contentDoc.documentURIObject, obj.value, true);
}


const PREF_COOKIE_BEHAVIOR = "network.cookie.cookieBehavior";

const Cookies = {
  _service: null,
  _prefs: null,

  start: function() {
    this._service = Cc["@mozilla.org/cookieService;1"].getService().QueryInterface(Ci.nsICookieService);
    this._prefs = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefBranch2);
    this._prefListener.behavior = this._prefs.getIntPref(PREF_COOKIE_BEHAVIOR);
    this._prefs.addObserver(PREF_COOKIE_BEHAVIOR, this._prefListener, false);
  },

  stop: function() {
    this._service = null;
    this._prefs.removeObserver(PREF_COOKIE_BEHAVIOR, this._prefListener);
    this._prefs = null;
  },


  _prefListener: {
    behavior: -1,

    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),
    // aTopic=nsPref:changed aData=network.cookie.cookieBehavior
    observe: function(aSubject, aTopic, aData) {
      this.behavior = aSubject
                      .QueryInterface(Ci.nsIPrefBranch)
                      .getIntPref(PREF_COOKIE_BEHAVIOR);
      console.log("pref! " + aSubject + aTopic + aData + this.behavior);
    }

  },

  setCookie: function(tabLogin, originalUri, originalCookie, fromJs) {
    var uri = toInternalUri(originalUri, tabLogin);
    var val = convertCookieDomain(originalCookie, tabLogin);

    if (this._prefListener.behavior === 0) {
      this._setCookie(fromJs, uri, val);
      return;
    }

    var p = this._prefs;
    p.removeObserver(PREF_COOKIE_BEHAVIOR, this._prefListener);
    p.setIntPref(PREF_COOKIE_BEHAVIOR, 0);
    this._setCookie(fromJs, uri, val);
    p.setIntPref(PREF_COOKIE_BEHAVIOR, this._prefListener.behavior);
    p.addObserver(PREF_COOKIE_BEHAVIOR, this._prefListener, false);
  },

  getCookie: function(fromJs, originalUri, tabLogin) {
    var uri = toInternalUri(originalUri, tabLogin);

    if (this._prefListener.behavior === 0) {
      return this._getCookie(fromJs, uri);
    }

    var p = this._prefs;
    p.removeObserver(PREF_COOKIE_BEHAVIOR, this._prefListener);
    p.setIntPref(PREF_COOKIE_BEHAVIOR, 0);
    var cookie = this._getCookie(fromJs, uri);
    p.setIntPref(PREF_COOKIE_BEHAVIOR, this._prefListener.behavior);
    p.addObserver(PREF_COOKIE_BEHAVIOR, this._prefListener, false);
    return cookie;
  },


  _setCookie: function(fromJs, uri, val) {
    if (fromJs) {
      this._service.setCookieString(uri,
                                    null,
                                    val,
                                    null);
    } else {
      //setCookieString doesn't work for httponly cookies
      this._service.setCookieStringFromHttp(uri,   // aURI
                                            null,  // aFirstURI
                                            null,  // aPrompt
                                            val,   // aCookie
                                            null,  // aServerTime
                                            null); // aChannel
    }
  },

  _getCookie: function(fromJs, uri) {
    if (fromJs) {
      return this._service.getCookieString(uri,
                                           null);
    } else {
      return this._service.getCookieStringFromHttp(uri,   // aURI
                                                   null,  // aFirstURI
                                                   null); // aChannel
    }
  }
};


function convertCookieDomain(cookieHeader, tabLogin) {
  var objCookies = new SetCookieParser(cookieHeader, true);
  var len = objCookies.length;
  var newCookies = new Array(len);

  for (var idx = 0; idx < len; idx++) {
    var myCookie = objCookies.getCookieByIndex(idx);
    var realDomain = myCookie.getStringProperty("domain");
    if (realDomain.length > 0) {
      myCookie.defineMeta("domain", toInternalHost(realDomain, tabLogin));
      newCookies[idx] = myCookie.toHeaderLine();
    } else {
      newCookies[idx] = myCookie.raw;
    }
  }

  return newCookies.join("\n");//objCookies.toHeader();
}


function toLines(txt) {
  return txt.split(/\r\n|\r|\n/);
}


function SetCookieParser(cookieHeader, isSetCookie) { // TODO remove isSetCookie
  this.m_hasMeta = isSetCookie;
  this._allCookies = [];
  if (cookieHeader !== null) {
    var lines = toLines(cookieHeader);
    var len = lines.length;
    if (isSetCookie) {
      for (var idx = 0; idx < len; idx++) {
        this.parseLineSetCookie(lines[idx]);
      }
    } else {
      for (var idx = 0; idx < len; idx++) {
        this.parseLineCookie(lines[idx]);
      }
    }
  }
}

SetCookieParser.prototype = {
  parseLineSetCookie: function(headerLine) {
    var unit = new CookieUnit(headerLine);
    var items = headerLine.split(";");

    for (var idx = 0, len = items.length; idx < len; idx++) {
      var pair = splitValueName(items[idx]);
      var name = pair[0];
      var value = pair[1]; // null ==> name=HttpOnly, secure etc

      if (idx === 0) {
        if (name.length > 0 && value !== null) {
          unit.defineValue(name, value);
        } else {
          util2.logEx("_allCookies invalid", name, value, headerLine);
          break;
        }
      } else {
        unit.defineMeta(name, value);
      }
    }

    this._allCookies.push(unit);
  },

  parseLineCookie: function(headerLine) {
    var items = headerLine.split(";");
    for (var idx = 0, len = items.length; idx < len; idx++) {
      var unit = CookieUnit(items[idx]);
      var pair = splitValueName(items[idx]);
      unit.defineValue(pair[0], pair[1]);
      this._allCookies.push(unit);
    }
  },

  /*
  toHeader: function() {
    var allCookies = this._allCookies;
    var len = allCookies.length;
    //var buf = [];
    var buf = new Array(len);
    for (var idx = 0; idx < len; idx++) {
      //if (allCookies[idx].value !== null) {
      buf[idx] = allCookies[idx].toHeaderLine();
      //}
    }
    return this.m_hasMeta ? buf.join("\n") : buf.join(";");
  },

  getCookie: function(name) {
    var aCookie = this._allCookies;
    for (var idx = 0, len = aCookie.length; idx < len; idx++) {
      if (name === aCookie[idx].name) {
        return aCookie[idx];
      }
    }
    return null;
  },

  forEach: function(fn) {
    var c = this._allCookies;
    for (var idx = 0, len = c.length; idx < len; idx++) {
      fn(c[idx]);
    }
  }
  */

  get length() {
    return this._allCookies.length;
  },

  getCookieByIndex: function(idx) {
    return this._allCookies[idx];
  }
};

function CookieUnit(line) {
  this._data = {//instance value
    "_src": line
  };
}

CookieUnit.prototype = {
  _data: null,

  clone: function() {
    var c = new CookieUnit();
    for (var n in this._data) {
      c._data[n] = this._data[n];
    }
    return c;
  },

  get name() {
    var rv = this._data["_name"];
    return rv ? rv : "";
  },

  get value() {
    var rv = this._data["_value"];
    return rv ? rv : "";
  },

  get raw() {
    return this._data["_src"];
  },

  defineValue: function(name, val) {
    this._data["_name"] = name;
    this._data["_value"] = val;
  },

  //"secure":
  //"httponly":
  hasBooleanProperty: function(name) {
    //name = name.toLowerCase();
    return name in this._data;
  },

  setBooleanProperty: function(name, def) {
    if (def) {
      this._data[name] = null;
    } else {
      delete this._data[name];
    }
  },

  //"expires":
  //"domain":
  //"path":
  getStringProperty: function(name) {
    var rv = this._data[name];
    return rv ? rv : "";
    //return rv || "";
  },

  defineMeta: function(name, val) {
    name = name.toLowerCase();
    switch (name) {
      case "expires":
      case "domain":
      case "path":
      case "secure":
      case "httponly":
        this._data[name] = val;
        break;
    }
  },

  toHeaderLine: function() {//toString()
    var buf = [this.name + "=" + this.value];
    var props;

    props = ["secure", "httponly"];
    for (var idx = 0, len = props.length; idx < len; idx++) {
      var propName = props[idx];
      if (this.hasBooleanProperty(propName)) {
        buf.push(propName.toUpperCase());
      }
    }

    props = ["expires", "path", "domain"];
    for (var idx = 0, len = props.length; idx < len; idx++) {
      var propName = props[idx];
      var val = this.getStringProperty(propName);
      if (val.length > 0) {
        buf.push(propName.toUpperCase() + "=" + val);
      }
    }

    return buf.join(";");
  }
};


function splitValueName(cookie) {
  var idx = cookie.indexOf("=");
  if (idx === -1) {
    return [cookie.trim(), null];
  }

  // "a=bcd=e".split("=",2) returns [a,bcd]
  //   "abcde".split("=",2) returns [abcde]


  // MY =
  // 012^-----idx=3 length=4

  // MY =a:1=6
  // 012^-----idx=3 length=9

  var pair = ["", ""];
  pair[0] = cookie.substring(0, idx).trim();
  idx++;
  if (idx < cookie.length) {
    pair[1] = cookie.substring(idx);
  }

  return pair;
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


function removeTldData_LS(tld) {
  // TODO del localStorage
}


function copyCookieToNewHost(cookie, newHost) {
  var expiryTime = cookie.isSession ? 4611686018427388000 : cookie.expiry; // Math.pow(2, 62)
  var mgr = Cc["@mozilla.org/cookiemanager;1"].getService(Ci.nsICookieManager2);
  mgr.add(newHost,         cookie.path,
          cookie.name,     cookie.value,
          cookie.isSecure, cookie.isHttpOnly, cookie.isSession, expiryTime);
}


function removeTldData_cookies(tld) {
  var all = getAllCookiesFromHost(tld);
  console.log('removeCookies n=', all.length);

  var mgr = Cc["@mozilla.org/cookiemanager;1"].getService(Ci.nsICookieManager2);
  var cookie;
  for (var idx = all.length - 1; idx > -1; idx--) {
    cookie = all[idx];
    mgr.remove(cookie.host, cookie.name, cookie.path, false);
  }
  return all;
}


// www.youtube.com.x-content.x-namespace.XXXXX.google.com.x-content.x-namespace

function getAllCookiesFromHost(h) {
  var rv = [];
  var mgr = Cc["@mozilla.org/cookiemanager;1"].getService(Ci.nsICookieManager2);
  var COOKIE = Ci.nsICookie2;

  var _qty = 0;
  var _t = new Date().getTime();

  // TODO use mgr.getCookiesFromHost when it doesn't have bugs
  var all = mgr.enumerator;
  while (all.hasMoreElements()) {
    var cookie = all.getNext().QueryInterface(COOKIE);
    _qty++;
    if (hasRootDomain(h, cookie.host)) {
      rv.push(cookie);
    }
  }

  console.log("getAllCookiesFromHost time=" + (new Date().getTime() - _t) + "ms -- len parsed=" + _qty + " -- len rv=" + rv.length);
  return rv;
}


function hasRootDomain(domain, host) {
  if (host === domain) {
    return true;
  }

  var idx = host.lastIndexOf(domain);
  if (idx === -1) {
    return false;
  }

  //return (index == (this.length - aDomain.length)) && (prevChar == "." || prevChar == "/");
  if ((host.length - domain.length) === idx) {
    return host[idx - 1] === ".";// || prev === "/";
  }

  return false;
}



function toInternalUri(uri, tabLogin) {
  if (tabLogin.loginUser16 === TabLogin.NewAccount) {
    return uri;
  }
  var u = uri.clone();
  u.host = toInternalHost(u.host, tabLogin);
  return u;
}

//getLoginFromInternalHost 1 .google.com.x-tab.x-session.jrklas.x-tab.x-session.google.com.x-tab.x-session4 - false
/*
multifox[04:06:42.337] getLoginFromInternalHost    .google.com.x-tab.x-session.jrklas.x-tab.x-session.google.com.x-tab.x-session
 [".google.com",
  ".jrklas",
  ".google.com",
  ""
]

multifox[04:06:42.337] getLoginFromInternalHost .statcounter.com.x-tab.x-session.gihfgmhegngbgogo.statcounter.com.x-tab.x-session
 [".statcounter.com",
  ".gihfgmhegngbgogo.statcounter.com",
  ""
]

*/
function getLoginFromInternalHost(host) {
  // host = ".twitter.com.<***>.abcdef.twitter.com.<***>"
  //         ====[0]=====      ========[1]========
  var arr = host.split(".x-content.x-namespace");
  //console.log('getLoginFromInternalHost',host, '\n',JSON.stringify(arr, null, 2));
  var len = arr.length;
  if ((len < 3) || (arr[len - 1] !== "")) {
    return null;
  }
  return {
    realHost:  arr[len - 3],
    fullLogin: arr[len - 2].substr(1) // arr[1]=".abcdefgh.login.tld"
  };
}


function toInternalHost(host, tabLogin) {
  if (tabLogin.loginUser16 === TabLogin.NewAccount) {
    console.log('toInternalHost TabLogin.NewAccount', host, tabLogin.loginTld, tabLogin.loginUser16);
    return host;
  }
  var SYMBOL = ".x-content.x-namespace";
  console.assert(typeof host === "string", "invalid domain2 ="+host);
  console.assert(typeof tabLogin.loginUser16 === "string", "invalid user=" + tabLogin.loginUser16);
  console.assert(typeof tabLogin.loginTld === "string", "invalid loginTld=" + tabLogin.loginTld);
  console.assert(tabLogin.loginUser16.indexOf(SYMBOL) === -1, "invalid user");
  return host + SYMBOL + "." + tabLogin.loginUser16 + "." + tabLogin.loginTld + SYMBOL;
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2010
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

function windowLocalStorage(obj, contentDoc) {
  if (contentDoc.documentURI === "about:blank") { // BUG 1.9.2
    console.log("windowLocalStorage doc=", contentDoc.documentURI, "/", contentDoc.location);
    return "";
  }

  var tabUser = TabLogin.getFromWindow(contentDoc.defaultView);
  console.assert(tabUser.loginTld !== null, "windowLocalStorage not logged");

  var uri = toInternalUri(contentDoc.documentURIObject, tabUser);
  var principal = Cc["@mozilla.org/scriptsecuritymanager;1"]
                  .getService(Ci.nsIScriptSecurityManager)
                  .getCodebasePrincipal(uri);

  var storage = Cc["@mozilla.org/dom/storagemanager;1"]
                .getService(Ci.nsIDOMStorageManager)
                .getLocalStorageForPrincipal(principal, "");

  var rv = undefined;
  switch (obj.cmd) {
    case "clear":
      storage.clear();
      break;
    case "removeItem":
      storage.removeItem(obj.key);
      break;
    case "setItem":
      storage.setItem(obj.key, obj.val); // BUG it's ignoring https
      break;
    case "getItem":
      rv = storage.getItem(obj.key);
      break;
    case "key":
      rv = storage.key(obj.index);
      break;
    case "length":
      rv = storage.length;
      break;
    default:
      throw new Error("localStorage interface unknown: " + obj.cmd);
  }

  console.log("localStorage " + uri.spec + "\n"+JSON.stringify(obj, null, 2) + "\n=====\nreturn " + rv);
  return rv;
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


const util2 = {

  logEx: function() {
    var buf = "logEx";// + this.logEx.caller + "\n";// + util2.throwStack._toString(false, Components.stack) + "===";
    for (var idx = 0, len = arguments.length; idx < len; idx++) {
      buf += "\n" + arguments[idx];// + "=" + this.caller[arguments[idx]];
    }

    buf += "\n====\n" + util2.throwStack._toString(false, Components.stack);
    console.log(buf);
  },

  throwStack: {

    go: function(txt) {
      console.log("Stack: " + txt + "\n" + this._toString(false, Components.stack));
      throw new Error(txt);
    },

    _toArray: function(webOnly, theStack) {
      var allItems = [];
      if (!theStack) {
        return allItems;
      }

      function StackItem() {}
      StackItem.prototype = {
        lang: "?",
        filename: "?",
        lineNumber: -1,
        name: "?"
      };

      // aparentemente é um bug que ocorre as vezes, todas as propriedades sao undefined
      if (theStack.languageName == undefined) {
        var item = new StackItem();
        item.name = theStack.toString();
        allItems.push(item);
      }

      // myStack.caller é quase sempre null, mas algumas vezes é undefined...
      for (var myStack = theStack; myStack; myStack = myStack.caller) {
        if (webOnly) {
          var n = myStack.filename;
          if (n === null) {
            continue;
          }

          if ((n.indexOf("http:") !== 0) && (n.indexOf("https:") !== 0)) {
            continue;
          }
        }


        var item = new StackItem();
        allItems.push(item);
        item.lang = myStack.languageName;
        item.filename = myStack.filename;
        item.lineNumber = myStack.lineNumber; // Valid line numbers begin at "1". "0" indicates unknown.
        item.name = myStack.name;
      }

      allItems.reverse();
      return allItems;
    },

    _toString: function(webOnly, theStack) {
      var arr = this._toArray(webOnly, theStack);
      var lines = new Array(arr.length);
      for (var idx = arr.length - 1, idx2 = 0; idx > -1; idx--) {
        var lang = webOnly ? "" : arr[idx].lang + " ";
        lines[idx2] = "[" + (idx + 1) + "] "
                    + lang
                    + arr[idx].filename + " "
                    + arr[idx].name + " ("
                    + arr[idx].lineNumber + ")";
        idx2++;
      }
      return lines.join("\n");
    }

  }
};


var TabLogin = {

  NewAccount: "",

  setLogin: function(tabLogin) {
    var tab = tabLogin.tabElement;
    if (tab.hasAttribute("multifox-tab-error")) {
      // reset error icon
      tab.removeAttribute("multifox-tab-error");
    }

    var domain = tabLogin.loginTld;
    var encodedUser = tabLogin.loginUser16;
    console.trace("setLogin " + encodedUser + " at " + domain);

    console.assert(domain !== null, "setLogin tld=null");
    console.assert(typeof(domain) === "string", "setLogin tld="+domain);
    console.assert(typeof(encodedUser) === "string", "setLogin user="+encodedUser);

    if (tab.hasAttribute("multifox-tab-login-tld") === false) {
      tab.setAttribute("multifox-logging-in", "true"); // TODO
    }

    tab.setAttribute("multifox-tab-login-tld", domain);
    tab.setAttribute("multifox-tab-login-user16", encodedUser);

    invalidateUI(tab);
  },


  setLoginInProgress: function(tab, username, domain) {
    tab.setAttribute("multifox-login-submit-domain", domain);
    tab.setAttribute("multifox-login-submit-user", username);
  },


  removeLogin: function(tab) {
    console.trace("removeLogin " + tab.getAttribute("multifox-tab-login-tld"));
    if (tab.hasAttribute("multifox-tab-login-tld")) {
      tab.removeAttribute("multifox-tab-login-tld");
      tab.removeAttribute("multifox-tab-login-user16");
      invalidateUI(tab);
    }
  },

  removeLoginInProgress: function(tab) {
    var moveMode = tab.getAttribute("multifox-login-submit-data-moved");
    tab.removeAttribute("multifox-login-submit-domain");
    tab.removeAttribute("multifox-login-submit-user");
    tab.removeAttribute("multifox-login-submit-data-moved");
    return moveMode;
  },

  hasLoginInProgressMoveData: function(tab) {
    return tab.hasAttribute("multifox-login-submit-data-moved");
  },

  setLoginInProgressMoveData: function(tab, mode) {
    tab.setAttribute("multifox-login-submit-data-moved", mode);
  },

  isLoginInProgress: function(tab) {
    return tab.hasAttribute("multifox-login-submit-domain");
  },

  getLoginInProgress: function(tab) {
    if (tab.hasAttribute("multifox-login-submit-domain")) { // BUG tab undef ao verificar update
      return {
        tabElement:  tab,
        loginUser16: stringToHex(tab.getAttribute("multifox-login-submit-user")),
        loginTld:    tab.getAttribute("multifox-login-submit-domain")
      }
    }
    return {
      tabElement:  tab,
      loginUser16: null,
      loginTld:    null
    }
  },


  getFromTab: function(tab) {
    if (tab.hasAttribute("multifox-tab-login-tld")) {
      return {
        tabElement:  tab,
        loginUser16: tab.getAttribute("multifox-tab-login-user16"),
        loginTld:    tab.getAttribute("multifox-tab-login-tld")
      }
    }

    return {
      tabElement:  tab,
      loginTld:    null,
      loginUser16: null
    }
  },


  getFromWindow: function(contentWin) {
    if (contentWin === null) {
      return {
        tabElement:  null,
        loginUser16: null,
        loginTld:    null
      };
    }

    var url = contentWin.document.documentURI;
    if (url.length === 0) {
      return { // TODO return null?
        tabElement:  null,
        loginUser16: null,
        loginTld:    null
      };
    }

    var tab = WindowParents.getContainerElement(contentWin);
    if (tab !== null) {
      return this.getFromTab(tab);
    }
    return {
      tabElement:  tab,
      loginUser16: null,
      loginTld:    null
    };
  }

};


const OpenerWindowIdentity = {

  findFromWindow: function(contentWin) {
    // popup via js/window.open?
    if (contentWin.opener) {
      //console.log("_getFromOpenerContent opener=" + contentWin.opener.document.location);
      var tabOpener = WindowParents.getContainerElement(contentWin.opener);
      if (tabOpener) {
        return TabLogin.getFromTab(tabOpener);
      }
    }

    //console.log("_getFromOpenerContent - no opener=" + contentWin.opener);

    // fx starting ==> opener=null
    return this.findFromOpenerSelectedTab(contentWin); // id from selected tab
  },


  findFromOpenerSelectedTab: function(contentWin) {
    var chromeWin = WindowParents.getChromeWindow(contentWin);
    if (chromeWin && chromeWin.opener) {
      var type = chromeWin.opener.document.documentElement.getAttribute("windowtype");
      if (type === "navigator:browser") {
        var selTab = chromeWin.opener.getBrowser().selectedTab;
        return TabLogin.getFromTab(selTab);
      }
    }
    return {
      tabElement:  null,
      loginUser16: null,
      loginTld:    null
    };

  }
};



const WindowParents = {
  getContainerElement: function(contentWin) {
    var chromeWin = this.getChromeWindow(contentWin);
    if ((chromeWin !== null) && ("getBrowser" in chromeWin)) {
      var elem = chromeWin.getBrowser();
      switch (elem.tagName) {
        case "tabbrowser":
          var topDoc = contentWin.top.document;
          var idx = elem.getBrowserIndexForDocument(topDoc);
          if (idx > -1) {
            return getTabs(elem)[idx];
          }
          break;
        default: // view-source => tagName="browser"
          // Note that this file uses document.documentURI to get
          // the URL (with the format from above). This is because
          // document.location.href gets the current URI off the docshell,
          // which is the URL displayed in the location bar, i.e.
          // the URI that the user attempted to load.
          console.log("getContainerElement=" + elem.tagName + "\n" +
                       contentWin.document.documentURI+ " " +chromeWin.document.location +"\n"+
                       contentWin.document.location   + " " +chromeWin.document.documentURI);
          break;
      }
    }
    return null;
  },

  getChromeWindow: function(contentWin) {
    if (!contentWin) console.trace('contentWin='+contentWin);
    var qi = contentWin.QueryInterface;
    if (!qi) {
      return null;
    }

    if (contentWin instanceof Ci.nsIDOMChromeWindow) {
      // extensions.xul, updates.xul ...
      return contentWin;
    }

    var win = qi(Ci.nsIInterfaceRequestor)
                .getInterface(Ci.nsIWebNavigation)
                .QueryInterface(Ci.nsIDocShell)
                .chromeEventHandler
                .ownerDocument
                .defaultView;
    // wrappedJSObject allows access to gBrowser etc
    // wrappedJSObject=undefined sometimes. e.g. contentWin=about:multifox
    var unwrapped = win.wrappedJSObject;
    return unwrapped ? unwrapped : win;
  }

};


function getTabs(tabbrowser) {
  return tabbrowser.mTabContainer.childNodes;
}


function stringToHex(str) {
  var alphabet = "abcdefghijklmnop";
  var len = str.length;
  var hex = new Array(len * 2);
  var j = 0;

  for (var idx = 0; idx < len; idx++) {
    var myByte = str.charCodeAt(idx);
    var nibble1 = (myByte & 0xf0) >> 4;
    var nibble0 =  myByte & 0x0f;
    hex[j++] = alphabet[nibble1];
    hex[j++] = alphabet[nibble0];
  }

  return hex.join("");
}


function hexToString(hex) {
  if (!hex) {
    console.trace("hexToString " + hex);
  }

  var offset = "a".charCodeAt(0); // "abcdefghijklmnop"
  var len = hex.length / 2;
  console.assert((len % 1) === 0, "invalid hex string: " + hex);
  var str = new Array(len);
  var j = 0;

  for (var idx = 0; idx < len; idx++) {
    var nibble1 = hex.charCodeAt(j++) - offset;
    var nibble0 = hex.charCodeAt(j++) - offset;
    str[idx] = String.fromCharCode((nibble1 << 4) | nibble0);
  }

  return str.join("");
}
